/***********************************************************************
|  Genealogy management system
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* textfile.c: manage text files */

#include <errno.h>
#ifdef MSDOS
#include <dos.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include "structs.h"

#define STARTENTRY 0x3fa58971L
#define HEADERSIZE sizeof(texthead)
#define READBUFSZ 2048


static struct {
long magic; /* magic number */
long end, limit; /* offsets, end of text and max end of text */
long d; /* date of text */
long source;
char type;
char filler[3];
} texthead;


/* get mod date, OS dependent */
long moddate(fd)
int fd; /* file descriptor */
{
long l;
#ifdef MSDOS
union REGS regs;

regs.h.ah = 0x57; /* get/set file date&time */
regs.h.al = 0;
regs.x.bx = fd;
intdos(&regs, &regs);
if(regs.x.cflag & 1)
eprint("could not get mod date for %d\n", fd);
l = (((long)(regs.x.dx)) << 16) + regs.x.cx;
#else
struct stat buf;

fstat(fd, &buf);
l = buf.st_mtime;
#endif

return l;
} /* moddate */

/* my own creat() routine.  Creates a binary file, rather than a text file
 * (under MSDOS), and checks for creation error. */
ecreat(fname)
char *fname;
{
return eopen(fname, O_BINARY|O_WRONLY|O_TRUNC|O_CREAT, 0666);
} /* creat */

ebinopen(fname)
char *fname;
{
return eopen(fname, O_RDONLY|O_BINARY);
} /* creat */

copyfile(source, dest, size)
long size;
{
int n;
char *s;

s = emalloc(READBUFSZ);
errno = 0;
while(size && !errno) {
n = read(source, s, READBUFSZ);
if(n > size) n = size;
write(dest, s, n);
size -= n;
}
free(s);
return errno;
} /* copyfile */

long archive_text(archname, archfd, outfd, ptr, textdate, texttype)
char *archname;
int archfd, outfd; /* file descriptors */
long ptr;
long textdate;
char texttype;
{
long archsize, textsize;
long ret_offset;
static char archfail[] = "cannot put text file back into archive %s[%ld, errno %d";

textsize = filesize(outfd);
elseek(outfd, 0L, 0);
if(ptr >= 0) {
elseek(archfd, ptr, 0);
eread(archfd, &texthead, HEADERSIZE);
}
archsize = filesize(archfd);
ret_offset = -1L;

if(textsize) {
/* text present, store it in the archive */
if(ptr >= 0 && ptr+HEADERSIZE + textsize <= texthead.limit) {
/* updated entry, and it fits in the old slot */
elseek(archfd, ptr, 0);
texthead.end = ptr+HEADERSIZE + textsize;
ewrite(archfd, &texthead, HEADERSIZE);
if(copyfile(outfd, archfd, textsize))
eprint(archfail, archname, ptr, errno);
ret_offset = ptr;
} else {

/* add this entry as new */
if(ptr >= 0) {
elseek(archfd, ptr, 0);
texthead.type = 'X';
ewrite(archfd, &texthead, HEADERSIZE);
}

elseek(archfd, archsize, 0);
texthead.magic = STARTENTRY;
texthead.end = archsize + textsize + HEADERSIZE;
texthead.limit = texthead.end + 16;
texthead.limit &= ~3L;
texthead.d = textdate;
texthead.source = this_dbid;
texthead.type = texttype;
texthead.filler[0] = texthead.filler[1] = texthead.filler[2] = 0;
ewrite(archfd, &texthead, HEADERSIZE);
if(copyfile(outfd, archfd, textsize))
eprint(archfail, archname, ptr, errno);
file_extend(archfd, texthead.limit, 0L);
ret_offset = archsize;
}
} /* text present */
else if(ptr >= 0) {
texthead.type = 'X';
elseek(archfd, ptr, 0);
ewrite(archfd, &texthead, HEADERSIZE);
}

close(outfd);
close(archfd);
return ret_offset;
} /* archive_text */

long textin_text(archname, textdate, texttype)
char *archname;
long textdate;
char texttype;
{
int archfd, outfd; /* file descriptors */

archfd = eopen(archname, O_BINARY|O_WRONLY|O_CREAT, 0666);
outfd = ebinopen(temp_textfile);
return archive_text(archname, archfd, outfd, -1L, textdate, texttype);
} /* textin_text */

user_edit()
{
char cmdstr[100];
char *s = getenv("ED");
if(!s || !*s)
#ifdef MSDOS
s = "edlin";
#else
s = "ed";
#endif
strncpy(cmdstr, s, 80);
cmdstr[80] = 0;
strcat(cmdstr,  " ");
strcat(cmdstr,  temp_textfile);
/* it's a tough call, but some editors return a nonzero exit status even
 * after a successful edit.  Run your version of ed,
 * make some changes, hit q (producing an error), then w then q.
 * Should be ok, but sometimes you get an exit status = 2.
 * Therefore we only abort on a negative return,
 * where the command could not be run. */
if(system(cmdstr) < 0)
eprint("error running system command:\n%s", cmdstr);
} /* user_edit */

long extract_file(archname, ptr)
char *archname;
long ptr;
{
int archfd, outfd; /* file descriptors */
long last_update, archsize, l;

outfd = ecreat(temp_textfile);
last_update = moddate(outfd);
archfd = open(archname, O_RDONLY|O_BINARY);
if(archfd < 0 && errno == EMFILE) {
uncache_fd();
archfd = open(archname, O_RDONLY|O_BINARY);
}
if(archfd < 0) { /* not present, create it */
archfd = ecreat(archname);
}
archsize = filesize(archfd);

if(ptr >= 0) {
if(ptr >= archsize)
eprint("offset %ld exceeds archive file size %ld", ptr, archsize);
elseek(archfd, ptr, 0);
eread(archfd, &texthead, HEADERSIZE);
l = ptr+HEADERSIZE;
if(texthead.magic != STARTENTRY)
eprint("bad magic number on start of text entry");

if(copyfile(archfd, outfd, texthead.end-l))
eprint("error extracting text entry from %s[%ld]", archname, ptr);
last_update = moddate(outfd);
} /* update existing text */
close(outfd);
close(archfd);
return last_update;
} /* extract_file */

long edfile(archname, ptr, op, textdate, texttype)
char *archname;
long ptr;
char op, texttype;
long textdate;
{
int archfd, outfd; /* file descriptors */
long last_update;

last_update = extract_file(archname, ptr);

user_edit();

/* now put the file back */
outfd = ebinopen(temp_textfile);
if(moddate(outfd) == last_update) goto bailout;
if(op == 'r' &&
!get_yn("the text has been changed, do you want to update this text entry?")) {
bailout:
close(outfd);
return ptr;
}

archfd = eopen(archname, O_RDWR|O_BINARY);
return archive_text(archname, archfd, outfd, ptr, textdate, texttype);
} /* edfile */

undos(fname)
char *fname;
{
#ifdef MSDOS
eprint("undos() called from MSDOS");
#else
int archfd, outfd; /* file descriptors */
char *s;
long archloc;
int size, n, i, j;
int count_nl, last_nl;

archfd = ebinopen(fname);
outfd = ecreat(temp_textfile);

archloc = 0;
while(eread(archfd, &texthead, HEADERSIZE)) { /* another entry */
count_nl = 0;
elseek(outfd, (long)HEADERSIZE, 1); /* we'll write the header later */
s = emalloc(READBUFSZ);
last_nl = 0;
size = texthead.end - archloc - HEADERSIZE;

while(size) {
n = size;
if(n > READBUFSZ) n = READBUFSZ;
n = eread(archfd, s, n);
size -= n;

/* compress cr/nl */
for(i=j=0; i<n; ++i) {
if(last_nl && s[i] == '\n') {
last_nl = 0;
++count_nl;
continue;
}
if(s[i] == '\r') {
last_nl = 1;
s[j++] = '\n';
continue;
}
s[j++] = s[i];
last_nl = 0;
} /* end crnl conversion */

ewrite(outfd, s, j);
} /* end massaging text file */
free(s);

/* now fix up header */
texthead.end -= count_nl;
elseek(outfd, archloc, 0);
ewrite(outfd, &texthead, HEADERSIZE);
archloc = texthead.limit;
file_extend(outfd, archloc, 0L);
elseek(archfd, archloc, 0);
} /* loop over text entries */

close(archfd);
close(outfd);
if(unlink(fname) || link(temp_textfile, fname) || unlink(temp_textfile))
eprint("cannot move UNDOSed temp file %s to %s", temp_textfile, fname);
#endif
} /* undos */

/* undos all text files */
mass_undos()
{
char *archname, *gng_nextfile();

dirstart = 1;
while(archname = gng_nextfile())
undos(archname);
} /* mass_undos */

setversion()
{
int vfd = ecreat("data/version");
long vnum = this_version;

#ifndef MSDOS
vnum |= 0x10000L;
#endif

ewrite(vfd, &vnum, 4);
close(vfd);
} /* setversion */

long getversion()
{
int vfd;
long vnum;

vfd = open("data/version", O_RDONLY|O_BINARY);
if(vfd < 0) {
if(errno != ENOENT)
eprint("cannot read data/version to establish the version of your database");
vnum = 0;
} else {
eread(vfd, &vnum, 4);
close(vfd);
}
return vnum;
} /* getversion */

checkversion()
{
int fd;
long vnum = getversion();

if(!vnum) {
clearscreen();
printf("This program builds and maintains your family tree in a genealogy database.\n\
It runs on all MSDOS machines and most UNIX machines.\n\
It may be freely distributed, as unmodified source, to anyone for\n\
personal non-profit purposes.\n");

if(get_yn("\nDo you wish instructions?"))
instruct();

/* the extra arg 0777 is harmless under MSDOS */
mkdir("data", 0777);
mkdir("idx", 0777);
mkdir("text", 0777);
dbidset();
serial_db_create();
setversion();
vnum = getversion();
setup_email();
}

/* read the unique database ID */
fd = ebinopen("data/dbid");
eread(fd, &this_dbid, 4);
close(fd);

version_upgrade(vnum);
} /* checkversion */

version_upgrade(vnum)
long vnum;
{
short unx, ver;

unx = vnum>>16;
ver = vnum;

if(ver <= 0 || ver > 1000) eprint("Database contains bogus version number %d", ver);

if(ver > this_version)
eprint("Your gng program is version %d, while Your database is version %d.\n\
Cannot access a more advanced database.");

if(ver == 1) { upgrade_1(); ++ver; }

#ifndef MSDOS
if(!unx) {
printf("Migrating from an MSDOS environment to a UNIX environment.\n\n");
mass_undos();
setup_email();
}
#endif
setversion();
} /* version_upgrade */

upgrade_1()
{
struct rel_cursor c;
struct address *a;
struct qknote qn;
int fd;

printf("Upgrading your database from version 1 to version 2, please wait...\n");

serial_rel_create(STRUCT_qknote);
fd = ecreat("idx/index_f");
close(fd);

init_0(&qn);
qn.ntype = 'b';

LOOP_address(&c, a, NULL)
qn.prsid = a->prsid;
strcpy(qn.note, a->free);
insupd_qknote(NULLKEY, &qn);
END_LOOP_address(&c)
} /* upgrade_1 */

/* display/dump instructions */
instruct()
{
char c;
char **s;
FILE *f;
extern char *inslines[];
int linecount;

linecount = 1;
nl();
for(s = inslines; *s; ++s) {
if(strcmp(*s, "-")) {
puts(*s);
++linecount;
continue;
}
while(linecount++ < 24) nl();
printf("\tHit any key to continue: ");
c = get_no_echo();
nl();
if(c == 'q') exit(1);
if(c == '.') { nl(); longjmp(main_menu, 2); }
nl();
linecount = 1;
}

while(linecount++ < 22) nl();
#ifdef MSDOS
printf("If you want a print copy of these instructions, set up your printer and type p.");
#endif
nl();
printf("If you want to save these instructions in the file `instruct', type s.\n");
printf("Otherwise hit any key to continue ");
c = get_no_echo();
nl();
nl();
if(c == 'q') exit(1);

#ifndef MSDOS
if(c == 'p') return;
#endif
if(c == 'p' || c == 's') {
f = efopen((c == 's') ? "instruct" : "lpt1", "w");
for(s = inslines; *s; ++s) {
if(!strcmp(*s, "-"))
fprintf(f, "\n");
else
fprintf(f, "%s\n", *s);
}
fclose(f);
} /* printing instructions */
} /* instruct */

/* set the (hopefully) unique database identification number */
dbidset()
{
long l;
int fd;

fd = ecreat("data/dbid");
l = moddate(fd);
ewrite(fd, &l, 4);
close(fd);
} /* dbidset */

unld_text(fd, f2, prsid, personal)
FILE *f2;
long prsid;
{
long offset = 0;
int i, cr;
unsigned int n;
long size;
char *s, c;

dot();

while(eread(fd, &texthead, HEADERSIZE)) {
if(texthead.type == 'p' || texthead.type == 'i' ||
personal && (texthead.type == 'c' || texthead.type == 'd')) {
offset += HEADERSIZE;
if(!texthead.d) s = "0/0/0";
else
s = datestring(texthead.d);
fprintf(f2, "<@*@>%ld.%ld%c%s\n", texthead.source, prsid, texthead.type, s);
size = texthead.end - offset;
if(!size) eprint("internal error, empty text entry");
s = emalloc(READBUFSZ);
cr = 0;
while(size) {
n = READBUFSZ;
if(n > size) n = size;
eread(fd, s, n);
size -= n;
for(i=0; i<n; ++i) {
c = s[i];
if(c&0x80)
eprint("unload/load feature only works with ASCII text entries");
#ifdef MSDOS /* cr nl crap */
if((c == '\n') ^ cr)
eprint("unload/load feature requires proper MSDOS CRLF sequences in text entries");
cr = 0;
if(c == '\r') { cr = 1; continue; }
#endif
eputc(c, f2);
}
} /* copying the entry */
if(c != '\n') eputc('\n', f2);

free(s);
}
offset = texthead.limit;
elseek(fd, offset, 0);
} /* loop over text entries */
} /* unld_text */

ld_text(f, foldin)
FILE *f;
{
int fd = -1;
int m, d, y;
int n;
long source, prsid;
long textdate;
long offset, offset1;
char etype;
long save_prsid = 0;
int save_y;
static char funny[] = "<@*@>";

if(!fgets(line, 256, f)) return;
if(strncmp(line, funny, 5))
eprint("badly formatted text entry header, should start with <@*@>");

top :
sscanf(line, "<@*@>%ld.%ld%c%d/%d/%d",
&source, &prsid, &etype, &m, &d, &y);
if(foldin) mapserial(&prsid);

if(prsid != save_prsid || y != save_y) {
if(fd >= 0) close(fd);
dot();
fd = ecreat(textfilename(prsid, 367L*y));
offset = 0;
save_prsid = prsid;
save_y = y;
}

texthead.magic = STARTENTRY;
texthead.source = source;
if(y) textdate = datecv(m,d,y);
else textdate = 0;
texthead.d = textdate;
texthead.type = etype;
texthead.filler[0] = texthead.filler[1] = texthead.filler[2] = 0;

elseek(fd, (long)HEADERSIZE, 1);
offset1 = offset + HEADERSIZE;

while(fgets(line, 256, f)) {
if(!strncmp(line, funny, 5)) break;
n = strlen(line);
ewrite(fd, line, n);
offset1 += n;
}

texthead.end = offset1;
offset1 += 16;
offset1 &= ~3L;
texthead.limit = offset1;
elseek(fd, offset, 0);
ewrite(fd, &texthead, HEADERSIZE);
file_extend(fd, offset1, 0L);

bloblink(prsid, textdate, offset, source, etype);
offset = offset1;
if(!strncmp(line, funny, 5)) goto top;

if(fd >= 0) close(fd);
} /* ld_text */

